#!/usr/bin/perl -w
#
# quizz CGI script
#
# Thomas Linden <tom@co.daemon.de>
#
# Copyright (c) 2003 Thomas Linden.
# ALL RIGHTS RESERVED.
#
# Licensed under the terms of the artistic license.
#
# The module "Template" is Copyright 1999-2000 Jens Heunemann.
# Used with permission.
#


use CGI qw/:standard/;
use Data::Dumper;
use CGI::Carp qw(fatalsToBrowser);
use CGI::Cookie;
use Template;
use Config::General;
use GDBM_File;
use Digest::MD5 qw(md5_base64);


# initialize vars, change as appropriate
my $cgi            = "index.cgi";
my $path_conf      = "config";
my $path_data      = "data";
my $path_templates = "templates";


# read/parse config file using Config::General
my %config = ParseConfig(
			 -ConfigFile       => "$path_conf/quizz.cfg",
			 -InterPolateVars  => 1,
			 -UseApacheInclude => 1,
			 -IncludeRelative  => 1
			   );

my %choice = (
	      M => "multiple",
	      O => "multiple",
	      S => "single",
	      R => "regex"
	      );

my $VERSION  = "1.1.5";

my @loop = @{$config{loop}};

# cgi object
my $query = new CGI;
my $now   = time;

# result+stats databases (gdbm)
my (%db, %stats) ;

# initialize if not already done so far
&prepare;

# initialize shared templates
my $head     = new Template("$path_templates/header.tpl");
my $foot     = new Template("$path_templates/footer.tpl");

# initialize cgi parameters
my %input = map { $_ => $query->param($_) } $query->param;

# check it up where to go
my $sub2go   = $input{state} || "default";




eval {
  # now do the real work
  &$sub2go;
};
if ($@) {
  print "Content-Type: text/html\n\n\n";
  print $@;
}
if ($config{debug}) {
  print "<pre>\n" . Dumper(\%input) . "\n</pre>";
}

# and the foot including the last changed date
my @stat = lstat($0);
$foot->code(
	    date => scalar localtime( $stat[9] ),
	    version => $VERSION
	   );
$foot->print;

exit;





###### all subs from now on, called by ::main via closure call (&$sub2go) #####
sub dump {
  my $var = shift;
  print "<pre>\n" . Dumper($var) . "\n</pre>";
}

sub stats {
  print "Content-Type: text/html\n\n\n";
  $head->print;
  my %e = &get_entry($input{user});
  my @stat = &get_stats;
  my $T = new Template(-file => "$path_templates/result.tpl");
  $T->code(
	   data => \%e,
	   form => { cgi => $cgi },
	   stat => \@stat
	  );
  print $T->process;
}

sub default {
  print "Content-Type: text/html\n\n\n";
  $head->print;
  my $T = new Template(-file => "$path_templates/default.tpl");

  # compute best-of list
  my @_bestof;
  foreach my $user (keys %db) {
    my %entry = &get_entry($user);
    push @_bestof, \%entry;
  }
  my $num = 1;
  foreach my $entry (sort { $b->{points} <=>  $a->{points} } @_bestof) {
    if ($num <= $config{config}->{max_best}) {
      $entry->{pos} = $num;
      $entry->{color} = $loop[0];
      @loop = reverse @loop;
      push @bestof, $entry;
    }
    else {
      last;
    }
    $num++;
  }

  $T->code(
	   data => {
		    quiz_name        => $config{quiz_name},
		    quiz_description => $config{quiz_description},
		    cgi              => $cgi
		   },
	   bestof => \@bestof
	  );

  print $T->process;
}




sub check_cookie {
  if ($config{config}->{cookie} ne "none") {
    # check cookie
    my %cookie = fetch CGI::Cookie;
    my ($name, $md5) = &get_md5;
    if (exists $cookie{$name}) {
      my $session = $cookie{$name}->value();
      if ($md5 == $session) {
	&default; # not allowed!
	print "<I>no retry as user <b>$input{user}</b> possible.</I>\n";
	return;
      }
    }
    else {
      # create new cookie
      my $c = new CGI::Cookie(
			      -name    => $name,
			      -value   => $md5,
			      -expires => $config{config}->{cookie_expire} || '+1M',   # expires after 1 week
			     );
      print header(-cookie=>[$c]);
      $head->print;
    }
  }
  else {
    # no cookies, but dont forget to send the header
    print "Content-Type: text/html\n\n\n";
    $head->print;
  }
  return 1;
}





sub compute_question {
  if (! exists $input{question}) {
    # starting point
    $input{question} = 1;

    # remove eventually existing stats entry for this user
    delete $stats{$input{user}};

    # check cookies and continue
    return if(! &check_cookie );
  }
  else {
    # previous run was a question
    print "Content-Type: text/html\n\n\n";
    $head->print;

    my %q = %{ $config{question}->{ $input{question} } }; # current question
    my $factor = $q{points} / $q{time};                   # how to calculate time from points
    my %entry = &get_entry($input{user});                 # results entry for current user, if any
    my ($points, $time, $result, %stats);

    if ($input{question} == 1) {
      # reset
      $entry{points} = 0;
      $entry{time}   = 0;
    }

    # calculate time and points based on time.
    # points will be zeroed if answer is wrong (later)!
    #
    # $points and $time               - summaries
    # $input{points} and $input{time} - current results
    $input{time}   = $now - $input{starttime};
    $input{points} = $q{points} - int($input{time} * $factor);
    $time          = $input{time} + $entry{time};
    $points        = $points = $input{points} + $entry{points};

    print "time needed: $input{time}<br>time required: $q{time}<br>" if($config{debug});

    if ($input{time} >= $q{time}) {
      # check for timeout
      $points = $entry{points};
      $input{points} = 0;
      $input{time}   = $now - $input{starttime};
      $result = "timeout";
    }
    else {
      # compute the answer
      $result = "correct"; # default
      if ($q{type} eq "M" || $q{type} eq "O") {
	# multiple/optional-multiple choice
	my %choices = map { $_ => 1 } $query->param("choice");
	my %correct = map { $_ => 1 } split /,/, $q{result};
	$input{given} = join ",", keys %choices;

	foreach my $choice (keys %choices) {
	  if (exists $correct{$choice}) {
	    delete $correct{$choice};
	    delete $choices{$choice};
	  }
	}
	my $remaining = scalar keys %correct; # not all possible were answered
	my $missing   = scalar keys %choices; # wrong answers were choosen

	if ($q{type} eq "M") {
	  if ($remaining || $missing) {
	    # shit, wrong answer!
	    $points = $entry{points};
	    $result = "wrong";
	  }
	}
	else {
	  # $q{type} eq "O"
	  # wrong if either a wrong answer were choosen
	  # or if not enough answers were given
	  if ($missing || $remaining >= $q{required}) {
	    # shit, wrong answer!
	    $points = $entry{points};
	    $result = "wrong";
	  }
	}
      }
      elsif ($q{type} eq "S") {
	# single choice
	$input{given} = $input{choice};
	if ($input{choice} ne $q{result}) {
	  # shit, wrong answer!
	  $points = $entry{points};
	  $result = "wrong";
	}
      }
      elsif ($q{type} eq "R") {
	# regex choice
	$input{given} = $input{choice};
	$input{given} =~ s/:/%%/g; # mask delimiter
	if ($input{choice} !~ /$q{result}/) {
	  # shit, wrong answer!
	  $points = $entry{points};
	  $result = "wrong";
	}
      }
    }

    print "result: $result<br>points: $input{points}" if($config{debug});

    &add_stats($input{question}, $input{given}, $input{time}, $input{points}, $result );
    &add_entry(time => $time, points => $points, user => $input{user}, date => scalar localtime(time));
    $input{question}++; # use next one
  }
  return 1;
}





sub quest {
  if ($input{user} eq "") {
    &default;
    print "<p>Username required</p>\n";
    return;
  }

  # check given answer, if any, and print cookie+content-type
  return if(! &compute_question );

  if (exists $config{question}->{ $input{question} }) {
    my $T = new Template(-file => "$path_templates/question.tpl");
    my %question = %{ $config{question}->{ $input{question} } };

    if ($config{config}->{counter_time}) {
      $T->code(
	       timer => {
			 max      => $question{time},
			 question => $question{question}
			}
	      );
    }

    if ($config{config}->{counter_points}) {
      $T->code(
	       pointer => {
			   max      => $question{points},
			   interval => $question{points} / $question{time},
			   question => $question{question}
			  }
	      );
    }

    # prepare choices
    my @answers;
    if ($question{type} eq "R") {
      my %c = ( choice => "Answer:");
      push @answers, \%c;
    }
    else {
      foreach my $choice (sort {$a cmp $b } keys %{$question{choice}}) {
	my %c = (
		 choice => $choice,
		 answer => $question{choice}->{$choice}
		);
	push @answers, \%c;
      }
    }

    $T->code(
	     $choice{ $question{type} } => \@answers,
	     form => {
		      question => $input{question},
		      user     => $input{user},
		      session  => $session,
		      cgi      => $cgi,
		      time     => time
		     }
	    );

    $T->code( data => { question => $input{question} } );

    print $T->process;
  }
  else {
    # no more questions, present the result
    my %e = &get_entry($input{user});
    my @stat = &get_stats;
    my $T = new Template(-file => "$path_templates/result.tpl");
    $T->code(
	     data => { %e},
	     stat => \@stat,
	     form => { cgi => $cgi }
	    );
    print $T->process;
  }
}








### helper subs ###
sub prepare {
  if (! -d "$path_data/$config{key}") {
    system "mkdir $path_data/$config{key}" and die "Could not prepare application!";
  }
  tie %db, "GDBM_File", "$path_data/$config{key}/results.db", &GDBM_WRCREAT, 0640
    or die "Could not open database $path_data/$config{key}/results.db: $!";
  tie %stats, "GDBM_File", "$path_data/$config{key}/stats.db", &GDBM_WRCREAT, 0640
    or die "Could not open database $path_data/$config{key}/stats.db: $!";
}

sub get_entry {
  my $user = shift;
  my ($points, $time, $date) = split /\//, $db{$user};
  my %entry = (
	       user   => $user,
	       date   => $date,
	       points => $points,
	       time   => $time
	      );
  return %entry;
}

sub add_entry {
  my %e = @_;
  $db{ $e{user} } = "$e{points}/$e{time}/$e{date}";
}


sub get_md5 {
  my ($name, $src);
  if ($config{config}->{cookie} eq "user") {
    $src  = $input{user};
    $name = "$input{user}-quizzsession";
  }
  elsif ($config{config}->{cookie} eq "ip") {
    $src  = $ENV{REMOTE_ADDR};
    $name = "$ENV{REMOTE_ADDR}-quizzsession";
  }
  elsif ($config{config}->{cookie} eq "both") {
    $input{user} . $ENV{REMOTE_ADDR};
    $name = "$input{user}-$ENV{REMOTE_ADDR}-quizzsession";
  }
  my $md5 = md5_base64($src);
  $q = $config{quiz_name};
  $q =~ s/\s//g;
  $name .= "-$q";
  return ($name, $md5);
}

sub get_stats {
  # format:
  # 1:A,C:20:380/..
  # question given-answers needed-time points
  my @entries = split /\//, $stats{$input{user}};
  my (%stat, @stat);
  foreach my $entry (@entries) {
    my($question, $given, $time, $points, $result) = split /:/, $entry;
    $given =~ s/%%/:/g; # de-mask delimiter
    if (!$given) {
      $given = "-";
    }
    my %q = %{ $config{question}->{$question} };
    my ($choice, $color);
    my %correct = map { $_ => 1 } split /,/, $q{result};
    if ($q{type} ne "R") {
      foreach my $answer (sort {$a cmp $b} keys %{$q{choice}}) {
	if (exists $correct{$answer}) {
	  $color = "green";
	}
	else {
	  $color = "red";
	}
	$choice .= "<b>$answer</b> - <font color=$color>$q{choice}->{$answer}</font><br>";
      }
    }
    else {
      $choice = $q{result};
    }
    if ($result eq "correct") {
      $color = "green";
    }
    else {
      $color = "red";
    }
    $stat{$question} = {
			pos      => $question,
			question => $q{question},
			given    => $given,
			choice   => $choice,
			time     => $time,
			points   => "$points/$q{points}",
			result   => "<font color=$color>$result</font>"
			};
  }

  foreach my $question (sort keys %stat) {
    push @stat, $stat{$question};
  }

  return @stat;
}

sub add_stats {
  # arguments:
  # ($question, $given, $time, $points, $result) = @_;
  # format:
  # 1:A,C:20:380/..
  # question given-answers needed-time points
  my $add  = join ":", @_;
  if (exists $stats{$input{user}}) {
    $add = "/" . $add;
  }
  $stats{$input{user}} .= $add;
}
